package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;

/** a parser which translates streams of Tokens of type T into a Forest<R> */
public abstract class Parser<Tok, Result> {

    protected final Table<Tok> pt;

    /** create a parser to parse the grammar with start symbol <tt>u</tt> */
    protected Parser(Union u, Topology<Tok> top)  { this.pt = new Table<Tok>(u, top); }
    protected Parser(Table<Tok> pt)               { this.pt = pt; }

    /** implement this method to create the output forest corresponding to a lone shifted input token */
    protected abstract Forest<Result> shiftToken(Input.Location oldloc, Tok t, Input.Location newloc);

    boolean helpgc = true;

    public String toString() { return pt.toString(); }

    /** parse <tt>input</tt>, using the table <tt>pt</tt> to drive the parser */
    public Forest<Result> parse(Input<Tok> input) throws IOException, ParseFailed {
        GSS gss = new GSS();
        Input.Location loc = input.getLocation();
        GSS.Phase current = gss.new Phase<Tok>(null, this, null, input.next(), loc, null);
        current.newNode(null, Forest.create(null, null, null, false), pt.start, true);
        int count = 1;
        for(int idx=0;;idx++) {
            Input.Location oldloc = loc;
            loc = input.getLocation();
            current.reduce();
            Forest forest = current.token==null ? null : shiftToken(oldloc, (Tok)current.token, loc);
            GSS.Phase next = gss.new Phase<Tok>(current, this, current, input.next(), loc, forest);
            if (!helpgc) {
                FileOutputStream fos = new FileOutputStream("out-"+idx+".dot");
                PrintWriter p = new PrintWriter(new OutputStreamWriter(fos));
                GraphViz gv = new GraphViz();
                for(Object n : next)
                    ((GSS.Phase.Node)n).toGraphViz(gv);
                gv.dump(p);
                p.flush();
                p.close();
            }
            count = next.size();
            if (current.isDone()) return (Forest<Result>)gss.finalResult;
            current = next;
        }
    }

    // Table //////////////////////////////////////////////////////////////////////////////

    /** an SLR(1) parse table which may contain conflicts */
    static class Table<Tok> extends Walk.Cache {

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("parse table");
            for(State<Tok> state : all_states.values()) {
                sb.append("  " + state + "\n");
                for(Topology<Tok> t : state.shifts) {
                    sb.append("      shift  \""+
                              new edu.berkeley.sbp.chr.CharTopology((IntegerTopology<Character>)t)+"\" => ");
                    for(State st : state.shifts.getAll(t))
                        sb.append(st.idx+"  ");
                    sb.append("\n");
                }
                for(Topology<Tok> t : state.reductions)
                    sb.append("      reduce \""+
                              new edu.berkeley.sbp.chr.CharTopology((IntegerTopology<Character>)t)+"\" => " +
                              state.reductions.getAll(t) + "\n");
            }
            return sb.toString();
        }

        public final Walk.Cache cache = this;

        private void walk(Element e, HashSet<Element> hs) {
            if (e==null) return;
            if (hs.contains(e)) return;
            hs.add(e);
            if (e instanceof Atom) return;
            for(Sequence s : (Union)e) {
                hs.add(s);
                for(Position p = s.firstp(); p != null; p = p.next())
                    walk(p.element(), hs);
            }
        }

        /** the start state */
        public final State<Tok>   start;

        /** used to generate unique values for State.idx */
        private int master_state_idx = 0;
        HashMap<HashSet<Position>,State<Tok>>   all_states    = new HashMap<HashSet<Position>,State<Tok>>();

        /** construct a parse table for the given grammar */
        public Table(Topology top) { this("s", top); }
        public Table(String startSymbol, Topology top) { this(new Union(startSymbol), top); }
        public Table(Union ux, Topology top) {
            Union start0 = new Union("0");
            start0.add(new Sequence.Singleton(ux));

            for(Sequence s : start0) cache.eof.put(s, true);
            cache.eof.put(start0, true);

            // construct the set of states
            HashSet<Element>                        all_elements  = new HashSet<Element>();
            walk(start0, all_elements);
            for(Element e : all_elements)
                cache.ys.addAll(e, new Walk.YieldSet(e, cache).walk());
            HashSet<Position> hp = new HashSet<Position>();
            reachable(start0, hp);
            this.start = new State<Tok>(hp, all_states, all_elements);

            // for each state, fill in the corresponding "row" of the parse table
            for(State<Tok> state : all_states.values())
                for(Position p : state.hs) {

                    // the Grammar's designated "last position" is the only accepting state
                    if (start0.contains(p.owner()) && p.next()==null)
                        state.accept = true;

                    if (isRightNullable(p)) {
                        Walk.Follow wf = new Walk.Follow(top.empty(), p.owner(), all_elements, cache);
                        Topology follow = wf.walk(p.owner());
                        for(Position p2 = p; p2 != null && p2.element() != null; p2 = p2.next())
                            follow = follow.intersect(new Walk.Follow(top.empty(), p2.element(), all_elements, cache).walk(p2.element()));
                        state.reductions.put(follow, p);
                        if (wf.includesEof()) state.eofReductions.add(p);
                    }

                    // if the element following this position is an atom, copy the corresponding
                    // set of rows out of the "master" goto table and into this state's shift table
                    if (p.element() != null && p.element() instanceof Atom)
                        state.shifts.addAll(state.gotoSetTerminals.subset(((Atom)p.element())));
                }
            if (top instanceof IntegerTopology)
                for(State<Tok> state : all_states.values()) {
                    state.oreductions = state.reductions.optimize(((IntegerTopology)top).functor());
                    state.oshifts = state.shifts.optimize(((IntegerTopology)top).functor());
                }
        }

        private boolean isRightNullable(Position p) {
            if (p.isLast()) return true;
            if (!possiblyEpsilon(p.element())) return false;
            return isRightNullable(p.next());
        }

        /** a single state in the LR table and the transitions possible from it */

        class State<Tok> implements Comparable<State<Tok>>, IntegerMappable, Iterable<Position> {
        
            public  final     int               idx    = master_state_idx++;
            private final     HashSet<Position> hs;

            public transient HashMap<Element,State<Tok>>          gotoSetNonTerminals = new HashMap<Element,State<Tok>>();
            private transient TopologicalBag<Tok,State<Tok>>     gotoSetTerminals    = new TopologicalBag<Tok,State<Tok>>();

            private           TopologicalBag<Tok,Position> reductions          = new TopologicalBag<Tok,Position>();
            private           HashSet<Position>              eofReductions       = new HashSet<Position>();
            private           TopologicalBag<Tok,State<Tok>>     shifts              = new TopologicalBag<Tok,State<Tok>>();
            private           boolean                         accept              = false;

            private VisitableMap<Tok,State<Tok>> oshifts = null;
            private VisitableMap<Tok,Position> oreductions = null;

            // Interface Methods //////////////////////////////////////////////////////////////////////////////

            boolean             isAccepting()               { return accept; }
            public Iterator<Position>  iterator()                  { return hs.iterator(); }

            boolean             canShift(Tok t)           { return oshifts.contains(t); }
            <B,C> void          invokeShifts(Tok t, Invokable<State<Tok>,B,C> irbc, B b, C c) {
                oshifts.invoke(t, irbc, b, c);
            }

            boolean             canReduce(Tok t)          { return t==null ? eofReductions.size()>0 : oreductions.contains(t); }
            <B,C> void          invokeReductions(Tok t, Invokable<Position,B,C> irbc, B b, C c) {
                if (t==null) for(Position r : eofReductions) irbc.invoke(r, b, c);
                else         oreductions.invoke(t, irbc, b, c);
            }

            // Constructor //////////////////////////////////////////////////////////////////////////////

            /**
             *  create a new state consisting of all the <tt>Position</tt>s in <tt>hs</tt>
             *  @param hs           the set of <tt>Position</tt>s comprising this <tt>State</tt>
             *  @param all_states   the set of states already constructed (to avoid recreating states)
             *  @param all_elements the set of all elements (Atom instances need not be included)
             *  
             *   In principle these two steps could be merged, but they
             *   are written separately to highlight these two facts:
             * <ul>
             * <li> Non-atom elements either match all-or-nothing, and do not overlap
             *      with each other (at least not in the sense of which element corresponds
             *      to the last reduction performed).  Therefore, in order to make sure we
             *      wind up with the smallest number of states and shifts, we wait until
             *      we've figured out all the token-to-position multimappings before creating
             *      any new states
             *  
             * <li> In order to be able to run the state-construction algorithm in a single
             *      shot (rather than repeating until no new items appear in any state set),
             *      we need to use the "yields" semantics rather than the "produces" semantics
             *      for non-Atom Elements.
             *  </ul>
             */
            public State(HashSet<Position> hs,
                         HashMap<HashSet<Position>,State<Tok>> all_states,
                         HashSet<Element> all_elements) {
                this.hs = hs;

                // register ourselves in the all_states hash so that no
                // two states are ever created with an identical position set
                all_states.put(hs, this);

                // Step 1a: examine all Position's in this state and compute the mappings from
                //          sets of follow tokens (tokens which could follow this position) to sets
                //          of _new_ positions (positions after shifting).  These mappings are
                //          collectively known as the _closure_

                TopologicalBag<Tok,Position> bag0 = new TopologicalBag<Tok,Position>();
                for(Position position : hs) {
                    if (position.isLast() || !(position.element() instanceof Atom)) continue;
                    Atom a = (Atom)position.element();
                    HashSet<Position> hp = new HashSet<Position>();
                    reachable(position.next(), hp);
                    bag0.addAll(a, hp);
                }

                // Step 1b: for each _minimal, contiguous_ set of characters having an identical next-position
                //          set, add that character set to the goto table (with the State corresponding to the
                //          computed next-position set).

                for(Topology<Tok> r : bag0) {
                    HashSet<Position> h = new HashSet<Position>();
                    for(Position p : bag0.getAll(r)) h.add(p);
                    gotoSetTerminals.put(r, all_states.get(h) == null ? new State<Tok>(h, all_states, all_elements) : all_states.get(h));
                }

                // Step 2: for every non-Atom element (ie every Element which has a corresponding reduction),
                //         compute the closure over every position in this set which is followed by a symbol
                //         which could yield the Element in question.
                //
                //         "yields" [in one or more step] is used instead of "produces" [in exactly one step]
                //         to avoid having to iteratively construct our set of States as shown in most
                //         expositions of the algorithm (ie "keep doing XYZ until things stop changing").
                HashMapBag<Element,Position> move = new HashMapBag<Element,Position>();
                for(Position p : hs) {
                    Element e = p.element();
                    if (e==null) continue;
                    for(Element y : cache.ys.getAll(e)) {
                        HashSet<Position> hp = new HashSet<Position>();
                        reachable(p.next(), hp);
                        move.addAll(y, hp);
                    }
                }
                for(Element y : move) {
                    HashSet<Position> h = move.getAll(y);
                    State<Tok> s = all_states.get(h) == null ? new State<Tok>(h, all_states, all_elements) : all_states.get(h);
                    gotoSetNonTerminals.put(y, s);
                }
            }

            public String toStringx() {
                StringBuffer st = new StringBuffer();
                for(Position p : this) {
                    if (st.length() > 0) st.append("\n");
                    st.append(p);
                }
                return st.toString();
            }
            public String toString() {
                StringBuffer ret = new StringBuffer();
                ret.append("state["+idx+"]: ");
                for(Position p : this) ret.append("{"+p+"}  ");
                return ret.toString();
            }

            public int compareTo(State<Tok> s) { return idx==s.idx ? 0 : idx < s.idx ? -1 : 1; }
            public int toInt() { return idx; }
        }
    }

    // Helpers //////////////////////////////////////////////////////////////////////////////
    
    private static void reachable(Element e, HashSet<Position> h) {
        if (e instanceof Atom) return;
        for(Sequence s : ((Union)e))
            reachable(s.firstp(), h);
    }
    private static void reachable(Position p, HashSet<Position> h) {
        if (h.contains(p)) return;
        h.add(p);
        if (p.element() != null) reachable(p.element(), h);
    }

}
